from dataclasses import field, dataclass

from kirin.ir import HasCFG, Method
from kirin.rewrite import (
    Walk,
    Chain,
    Fixpoint,
    Call2Invoke,
    ConstantFold,
    CFGCompactify,
    InlineGetItem,
    DeadCodeElimination,
)
from kirin.passes.abc import Pass
from kirin.rewrite.abc import RewriteResult

from .hint_const import HintConst


@dataclass
class Fold(Pass):
    """
    Pass that runs a number of small optimization rewrites.

    Specifically, the following rewrites are chained:

    - `ConstantFold`
    - `InlineGetItem`
    - `Call2Invoke`
    - `DeadCodeElimination`
    """

    hint_const: HintConst = field(init=False)

    def __post_init__(self):
        self.hint_const = HintConst(self.dialects)
        self.hint_const.no_raise = self.no_raise

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = self.hint_const.unsafe_run(mt)
        result = (
            Fixpoint(
                Walk(
                    Chain(
                        ConstantFold(),
                        InlineGetItem(),
                        Call2Invoke(),
                        DeadCodeElimination(),
                    )
                )
            )
            .rewrite(mt.code)
            .join(result)
        )

        if mt.code.has_trait(HasCFG):
            result = Walk(CFGCompactify()).rewrite(mt.code).join(result)

        return Fixpoint(Walk(DeadCodeElimination())).rewrite(mt.code).join(result)



"""Bindings for built-in types."""

import numbers

from kirin.ir.attrs.types import (
    Union as Union,
    Vararg as Vararg,
    AnyType as AnyType,
    Generic as Generic,
    Literal as Literal,
    PyClass as PyClass,
    TypeVar as TypeVar,
    BottomType as BottomType,
    FunctionType as FunctionType,
    TypeAttribute as TypeAttribute,
    TypeofMethodType as TypeofMethodType,
    hint2type as hint2type,
    is_tuple_of as is_tuple_of,
)

Any = AnyType()
Bottom = BottomType()
Int = PyClass(int)
Float = PyClass(float)
Complex = PyClass(complex)
Number = PyClass(numbers.Number)
String = PyClass(str)
Bool = PyClass(bool)
NoneType = PyClass(type(None))
List = Generic(list, TypeVar("T"))
Slice = Generic(slice, TypeVar("T"))
Tuple = Generic(tuple, Vararg(TypeVar("T")))
Dict = Generic(dict, TypeVar("K"), TypeVar("V"))
Set = Generic(set, TypeVar("T"))
FrozenSet = Generic(frozenset, TypeVar("T"))
MethodType = TypeofMethodType()
